# coding: utf8

from __future__ import unicode_literals
import codecs
from .utils import maketrans

buckwalter_transliteration = maketrans("'>&<}AbptvjHxd*rzs$SDTZEg_fqklmnhwYyFNKaui~o^#`{:@\"[;,.!-+%]", "\u0621\u0623\u0624\u0625\u0626\u0627\u0628\u0629\u062a\u062b\u062c\u062d\u062e\u062f\u0630\u0631\u0632\u0633\u0634\u0635\u0636\u0637\u0638\u0639\u063a\u0640\u0641\u0642\u0643\u0644\u0645\u0646\u0647\u0648\u0649\u064a\u064b\u064c\u064d\u064e\u064f\u0650\u0651\u0652\u0653\u0654\u0670\u0671\u06dc\u06df\u06e0\u06e2\u06e3\u06e5\u06e6\u06e8\u06ea\u06eb\u06ec\u06ed")


class QuranCorpusReader():
	"""
	interfaces [Quran Corpus](http://corpus.quran.com/download/) that you must download and extract it.

	>>> quran = QuranCorpusReader(quran_file='corpora/quranic-corpus-morphology.txt')
	>>> print(next(quran.words())[1])
	بِسْمِ
	"""

	def __init__(self, quran_file):
		self._quran_file = quran_file

	def parts(self):
		for line in codecs.open(self._quran_file):
			if not line.startswith('('):
				continue
			parts = line.strip().split('\t')

			part = {'loc': eval(parts[0].replace(':', ',')), 'text': parts[1].translate(buckwalter_transliteration), 'tag': parts[2]}

			features = parts[3].split('|')
			for feature in features:
				if feature.startswith('LEM:'):
					part['lem'] = feature[4:].translate(buckwalter_transliteration)
				elif feature.startswith('ROOT:'):
					part['root'] = feature[5:].translate(buckwalter_transliteration)
			yield part

	def words(self):

		def word_item(location, parts):
			text = ''.join([part['text'] for part in parts])
			tag = '-'.join([part['tag'] for part in parts])
			lem = '-'.join([part['lem'] for part in parts if 'lem' in part])
			root = '-'.join([part['root'] for part in parts if 'root' in part])
			return '.'.join(map(str, location)), text, lem, root, tag, parts

		last_location = (0, 0, 0, 0)
		items = []
		for part in self.parts():
			if last_location[:3] == part['loc'][:3]:
				items.append(part)
			else:
				if items:
					yield word_item(last_location[:3], items)
				items = [part]
			last_location = part['loc']
			del part['loc']
		yield word_item(last_location[:3], items)



# coding: utf-8

from __future__ import unicode_literals
import codecs
from collections import namedtuple


Verb = namedtuple('Verb', ('past_light_verb', 'present_light_verb', 'prefix', 'nonverbal_element', 'preposition', 'valency'))


class VerbValencyReader():
	"""
	interfaces [Verb Valency Corpus](http://dadegan.ir/catalog/pervallex)
	Mohammad Sadegh Rasooli, Amirsaeid Moloodi, Manouchehr Kouhestani, & Behrouz Minaei Bidgoli. (2011). A Syntactic Valency Lexicon for Persian Verbs: The First Steps towards Persian Dependency Treebank. in 5th Language & Technology Conference (LTC): Human Language Technologies as a Challenge for Computer Science and Linguistics (pp. 227–231). Poznań, Poland.
	"""

	def __init__(self, valency_file='corpora/valency.txt'):
		self._valency_file = valency_file

	def verbs(self):
		with codecs.open(self._valency_file, encoding='utf-8') as valency_file:
			for line in valency_file:
				if 'بن ماضی' in line:
					continue

				line = line.strip().replace('-\t', '\t')
				parts = line.split('\t')
				if len(parts) == 6:
					yield Verb(*parts)



# coding: utf-8

from __future__ import print_function, unicode_literals
import os, codecs, tempfile
from nltk.parse import DependencyGraph
from nltk.parse.api import ParserI
from nltk.parse.malt import MaltParser


class MaltParser(MaltParser):
	"""
	interfaces [MaltParser](http://www.maltparser.org/)
	"""

	def __init__(self, tagger, lemmatizer, working_dir='resources', model_file='langModel.mco'):
		self.tagger = tagger
		self.working_dir = working_dir
		self.mco = model_file
		self._malt_bin = os.path.join(working_dir, 'malt.jar')
		self.lemmatize = lemmatizer.lemmatize if lemmatizer else lambda w, t: '_'

	def parse_sents(self, sentences, verbose=False):
		tagged_sentences = self.tagger.tag_sents(sentences)
		return self.parse_tagged_sents(tagged_sentences, verbose)

	def parse_tagged_sents(self, sentences, verbose=False):
		input_file = tempfile.NamedTemporaryFile(prefix='malt_input.conll', dir=self.working_dir, delete=False)
		output_file = tempfile.NamedTemporaryFile(prefix='malt_output.conll', dir=self.working_dir, delete=False)

		try:
			for sentence in sentences:
				for i, (word, tag) in enumerate(sentence, start=1):
					word = word.strip()
					if not word:
						word = '_'
					input_file.write(('\t'.join([str(i), word.replace(' ', '_'), self.lemmatize(word, tag).replace(' ', '_'), tag, tag, '_', '0', 'ROOT', '_', '_', '\n'])).encode('utf8'))
				input_file.write('\n\n'.encode('utf8'))
			input_file.close()

			cmd = ['java', '-jar', self._malt_bin, '-w', self.working_dir, '-c', self.mco, '-i', input_file.name, '-o', output_file.name, '-m', 'parse']
			if self._execute(cmd, verbose) != 0:
				raise Exception("MaltParser parsing failed: %s" % (' '.join(cmd)))

			return (DependencyGraph(item) for item in codecs.open(output_file.name, encoding='utf8').read().split('\n\n') if item.strip())

		finally:
			input_file.close()
			os.remove(input_file.name)
			output_file.close()
			os.remove(output_file.name)


class TurboParser(ParserI):
	"""
	interfaces [TurboParser](http://www.ark.cs.cmu.edu/TurboParser/) which you must manually install
	"""

	def __init__(self, tagger, lemmatizer, model_file):
		self.tagger = tagger
		self.lemmatize = lemmatizer.lemmatize if lemmatizer else lambda w, t: '_'

		import turboparser
		self._pturboparser = turboparser.PTurboParser()
		self.interface = self._pturboparser.create_parser()
		self.interface.load_parser_model(model_file)

	def parse_sents(self, sentences):
		tagged_sentences = self.tagger.tag_sents(sentences)
		return self.tagged_parse_sents(tagged_sentences)

	def tagged_parse_sents(self, sentences):
		input_file = tempfile.NamedTemporaryFile(prefix='turbo_input.conll', dir='resources', delete=False)
		output_file = tempfile.NamedTemporaryFile(prefix='turbo_output.conll', dir='resources', delete=False)

		try:
			for sentence in sentences:
				for i, (word, tag) in enumerate(sentence, start=1):
					word = word.strip()
					if not word:
						word = '_'
					input_file.write(('\t'.join([str(i), word.replace(' ', '_'), self.lemmatize(word, tag).replace(' ', '_'), tag, tag, '_', '0', 'ROOT', '_', '_', '\n'])).encode('utf8'))
				input_file.write('\n'.encode('utf8'))
			input_file.close()

			self.interface.parse(input_file.name, output_file.name)

			return (DependencyGraph(item, cell_extractor=lambda cells: cells[1:8]) for item in codecs.open(output_file.name, encoding='utf8').read().split('\n\n') if item.strip())

		finally:
			input_file.close()
			os.remove(input_file.name)
			output_file.close()
			os.remove(output_file.name)


class DependencyParser(MaltParser):
	"""
	>>> from hazm import POSTagger, Lemmatizer
	>>> parser = DependencyParser(tagger=POSTagger(model='resources/postagger.model'), lemmatizer=Lemmatizer())
	>>> parser.parse(['من', 'به', 'مدرسه', 'رفته بودم', '.']).tree().pprint()
	(رفته_بودم من (به مدرسه) .)
	"""



# coding: utf-8

from __future__ import unicode_literals, print_function
import os, sys, itertools
from xml.dom import minidom


class SentiPersReader():
	"""
	interfaces [SentiPers Corpus](http://nlp.guilan.ac.ir/Dataset.aspx)

	>>> sentipers = SentiPersReader(root='corpora/sentipers')
	>>> next(sentipers.comments())[0][1]
	'بيشتر مناسب است براي کساني که به دنبال تنوع هستند و در همه چيز نو گرايي دارند .'
	"""

	def __init__(self, root):
		self._root = root

	def docs(self):

		def element_sentences(element):
			for sentence in element.getElementsByTagName('Sentence'):
				yield {'text': sentence.childNodes[0].data, 'id': sentence.getAttribute('ID'), 'value': int(sentence.getAttribute('Value')) if comment.getAttribute('Value') else None}

		for root, dirs, files in os.walk(self._root):
			for filename in sorted(files):
				try:
					elements = minidom.parse(os.path.join(root, filename))

					product = elements.getElementsByTagName('Product')[0]
					doc = {
						'Title': product.getAttribute('Title'),
						'Type': product.getAttribute('Type'),
						'comments': [],
					}

					for child in product.childNodes:
						if child.nodeName in {'Voters', 'Performance', 'Capability', 'Production_Quality', 'Ergonomics', 'Purchase_Value'}:
							value = child.getAttribute('Value')
							doc[child.nodeName] = float(value) if '.' in value else int(value)

					for comment in itertools.chain(elements.getElementsByTagName('Opinion'), elements.getElementsByTagName('Criticism')):
						doc['comments'].append({
							'id': comment.getAttribute('ID'),
							'type': comment.nodeName,
							'author': comment.getAttribute('Holder').strip(),
							'value': int(comment.getAttribute('Value')) if comment.getAttribute('Value') else None,
							'sentences': list(element_sentences(comment))
						})

					# todo: Accessories, Features, Review, Advantages, Tags, Keywords, Index

					yield doc

				except Exception as e:
					print('error in reading', filename, e, file=sys.stderr)

	def comments(self):
		for doc in self.docs():
			yield [[sentence['text'] for sentence in text] for text in [comment['sentences'] for comment in doc['comments']]]



# coding: utf-8

from __future__ import unicode_literals
import re
from nltk.tokenize.api import TokenizerI


class SentenceTokenizer(TokenizerI):
	"""
	>>> tokenizer = SentenceTokenizer()
	>>> tokenizer.tokenize('جدا کردن ساده است. تقریبا البته!')
	['جدا کردن ساده است.', 'تقریبا البته!']
	"""

	def __init__(self):
		self.pattern = re.compile(r'([!\.\?⸮؟]+)[ \n]+')

	def tokenize(self, text):
		text = self.pattern.sub(r'\1\n\n', text)
		return [sentence.replace('\n', ' ').strip() for sentence in text.split('\n\n') if sentence.strip()]



# coding: utf-8

from __future__ import unicode_literals
import re, codecs
from .Normalizer import *
from .PeykareReader import join_verb_parts

default_pos_map = {'ADJ': 'ADJ', 'ADJ_CMPR': 'ADJ', 'ADJ_INO': 'ADJ', 'ADJ_ORD': 'ADJ', 'ADJ_SIM': 'ADJ', 'ADJ_SUP': 'ADJ', 'ADV': 'ADV', 'ADV_EXM': 'ADV', 'ADV_I': 'ADV', 'ADV_NEGG': 'ADV', 'ADV_NI': 'ADV', 'ADV_TIME': 'ADV', 'AR': 'AR', 'CON': 'CONJ', 'DEFAULT': 'DEFAULT', 'DELM': 'PUNC', 'DET': 'PREP', 'IF': 'IF', 'INT': 'INT', 'MORP': 'MORP', 'MQUA': 'MQUA', 'MS': 'MS', 'N_PL': 'N', 'N_SING': 'N', 'NN': 'NN', 'NP': 'NP', 'OH': 'OH', 'OHH': 'OHH', 'P': 'PREP', 'PP': 'PP', 'PRO': 'PR', 'PS': 'PS', 'QUA': 'QUA', 'SPEC': 'SPEC', 'V_AUX': 'V', 'V_IMP': 'V', 'V_PA': 'V', 'V_PRE': 'V', 'V_PRS': 'V', 'V_SUB': 'V'}


class BijankhanReader():
	"""
	interfaces [Bijankhan Corpus](http://ece.ut.ac.ir/dbrg/bijankhan/Corpus/BijanKhan_Corpus_Processed.zip) that you must download and extract it.

	>>> bijankhan = BijankhanReader(bijankhan_file='corpora/bijankhan.txt')
	>>> next(bijankhan.sents())
	[('اولین', 'ADJ'), ('سیاره', 'N'), ('خارج', 'ADJ'), ('از', 'PREP'), ('منظومه', 'N'), ('شمسی', 'ADJ'), ('دیده_شد', 'V'), ('.', 'PUNC')]
	"""

	def __init__(self, bijankhan_file, joined_verb_parts=True, pos_map=default_pos_map):
		self._bijankhan_file = bijankhan_file
		self._joined_verb_parts = joined_verb_parts
		self._pos_map = pos_map
		self._normalizer = Normalizer(punctuation_spacing=False)

	def _sentences(self):
		sentence = []
		for line in codecs.open(self._bijankhan_file, encoding='utf-8'):
			parts = re.split('  +', line.strip())
			if len(parts) == 2:
				word, tag = parts
				if word not in ('#', '*'):
					word = self._normalizer.normalize(word)
					sentence.append((word if word else '_', tag))
				if tag == 'DELM' and word in ('#', '*', '.', '؟', '!') :
					if len(sentence):
						yield sentence
						sentence = []

	def sents(self):
		map_poses = lambda item: (item[0], self._pos_map.get(item[1], item[1]))

		for sentence in self._sentences():
			if self._joined_verb_parts:
				sentence = join_verb_parts(sentence)

			yield list(map(map_poses, sentence))




from .WordTokenizer import WordTokenizer
from .SentenceTokenizer import SentenceTokenizer
from .TokenSplitter import TokenSplitter
from .HamshahriReader import HamshahriReader
from .PersicaReader import PersicaReader
from .BijankhanReader import BijankhanReader
from .PeykareReader import PeykareReader
from .VerbValencyReader import VerbValencyReader
from .DadeganReader import DadeganReader
from .TreebankReader import TreebankReader
from .WikipediaReader import WikipediaReader
from .SentiPersReader import SentiPersReader
from .QuranCorpusReader import QuranCorpusReader
from .TNewsReader import TNewsReader
from .Normalizer import Normalizer
from .InformalNormalizer import InformalNormalizer, InformalLemmatizer
from .Stemmer import Stemmer
from .Lemmatizer import Lemmatizer
from .SequenceTagger import SequenceTagger, IOBTagger
from .POSTagger import POSTagger, StanfordPOSTagger
from .Chunker import Chunker, RuleBasedChunker, tree2brackets
from .DependencyParser import DependencyParser, MaltParser, TurboParser


from .utils import words_list, stopwords_list


def sent_tokenize(text):
	if not hasattr(sent_tokenize, 'tokenizer'):
		sent_tokenize.tokenizer = SentenceTokenizer()
	return sent_tokenize.tokenizer.tokenize(text)


def word_tokenize(sentence):
	if not hasattr(word_tokenize, 'tokenizer'):
		word_tokenize.tokenizer = WordTokenizer()
	return word_tokenize.tokenizer.tokenize(sentence)



# coding: utf-8

from __future__ import unicode_literals
from nltk.chunk import ChunkParserI, RegexpParser, tree2conlltags, conlltags2tree
from .SequenceTagger import IOBTagger


def tree2brackets(tree):
	str, tag = '', ''
	for item in tree2conlltags(tree):
		if item[2][0] in {'B', 'O'} and tag:
			str += tag +'] '
			tag = ''

		if item[2][0] == 'B':
			tag = item[2].split('-')[1]
			str += '['
		str += item[0] +' '

	if tag:
		str += tag +'] '

	return str.strip()


class Chunker(IOBTagger, ChunkParserI):
	"""
	>>> chunker = Chunker(model='resources/chunker.model')
	>>> tree2brackets(chunker.parse([('نامه', 'Ne'), ('ایشان', 'PRO'), ('را', 'POSTP'), ('دریافت', 'N'), ('داشتم', 'V'), ('.', 'PUNC')]))
	'[نامه ایشان NP] [را POSTP] [دریافت داشتم VP] .'
	"""

	def train(self, trees):
		super(Chunker, self).train(map(tree2conlltags, trees))

	def parse(self, sentence):
		return next(self.parse_sents([sentence]))

	def parse_sents(self, sentences):
		for conlltagged in super(Chunker, self).tag_sents(sentences):
			yield conlltags2tree(conlltagged)

	def evaluate(self, gold):
		return ChunkParserI.evaluate(self, gold)


class RuleBasedChunker(RegexpParser):
	"""
	>>> chunker = RuleBasedChunker()
	>>> tree2brackets(chunker.parse([('نامه', 'Ne'), ('۱۰', 'NUMe'), ('فوریه', 'Ne'), ('شما', 'PRO'), ('را', 'POSTP'), ('دریافت', 'N'), ('داشتم', 'V'), ('.', 'PUNC')]))
	'[نامه ۱۰ فوریه شما NP] [را POSTP] [دریافت داشتم VP] .'
	"""

	def __init__(self):
		grammar = r"""

			NP:
				<P>{<N>}<V>

			VP:
				<.*[^e]>{<N>?<V>}
				{<V>}

			ADVP:
				{<ADVe?><AJ>?}

			ADJP:
				<.*[^e]>{<AJe?>}

			NP:
				{<DETe?|Ne?|NUMe?|AJe|PRO|CL|RESe?><DETe?|Ne?|NUMe?|AJe?|PRO|CL|RESe?>*}
				<N>}{<.*e?>

			ADJP:
				{<AJe?>}

			POSTP:
				{<POSTP>}

			PP:
				{<Pe?>+}

		"""

		super(RuleBasedChunker, self).__init__(grammar=grammar)



# coding: utf-8

from __future__ import unicode_literals
import codecs
from .utils import informal_verbs, informal_words, NUMBERS
from .Normalizer import Normalizer
from .Lemmatizer import Lemmatizer
from .Stemmer import Stemmer
from .WordTokenizer import *
from .SentenceTokenizer import *


class InformalNormalizer(Normalizer):

	def __init__(self, verb_file=informal_verbs, word_file=informal_words, seperation_flag=False, **kargs):
		self.seperation_flag = seperation_flag
		self.lemmatizer = Lemmatizer()
		self.ilemmatizer = InformalLemmatizer()
		self.stemmer = Stemmer()
		super(InformalNormalizer, self).__init__(**kargs)

		def informal_to_formal_conjucation(i, f, flag):
			iv = self.informal_conjugations(i)
			fv = self.lemmatizer.conjugations(f)
			res = {}
			if flag:
				for i, j in zip(iv, fv[48:]):
					res[i] = j
					if '‌' in i:
						res[i.replace('‌', '')] = j
						res[i.replace('‌', ' ')] = j
					if i.endswith('ین'):
						res[i[:-1] + 'د'] = j
			else:
				for i, j in zip(iv[8:], fv[56:]):
					res[i] = j
					if '‌' in i:
						res[i.replace('‌', '')] = j
						res[i.replace('‌', ' ')] = j
					if i.endswith('ین'):
						res[i[:-1] + 'د'] = j

			return res

		with codecs.open(verb_file, encoding='utf8') as vf:
			self.iverb_map = {}
			for f, i, flag in map(lambda x: x.strip().split(' ', 2), vf):
				self.iverb_map.update(
					informal_to_formal_conjucation(i, f, flag)
				)

		with codecs.open(word_file, encoding='utf8') as wf:
			self.iword_map = dict(
				map(lambda x: x.strip().split(' ', 1), wf)
			)

		self.words = set()
		if self.seperation_flag:
			self.words.update(self.iword_map.keys())
			self.words.update(self.iword_map.values())
			self.words.update(self.iverb_map.keys())
			self.words.update(self.iverb_map.values())
			self.words.update(self.lemmatizer.words)
			self.words.update(self.lemmatizer.verbs.keys())
			self.words.update(self.lemmatizer.verbs.values())

	def split_token_words(self, token):

		def shekan(token):
			res = ['']
			for i in token:
				res[-1] += i
				if i in set(['ا', 'د', 'ذ', 'ر', 'ز', 'ژ', 'و'] + list(NUMBERS)):
					res.append('')
			while '' in res:
				res.remove('')
			return res

		def perm(lst):
			if len(lst) > 1:
				up = perm(lst[1:])
			else:
				return [lst]
			res = []
			for i in up:
				res.append([lst[0]] + i)
				res.append([lst[0] + i[0]] + i[1:])
			res.sort(key=len)
			return res

		token = re.sub(r'(.)\1{2,}', r'\1', token)
		ps = perm(shekan(token))
		for c in ps:
			if set(map(lambda x: self.ilemmatizer.lemmatize(x), c)).issubset(self.words):
				return ' '.join(c)
		return token


	def normalized_word(self, word):
		"""
		>>> normalizer = InformalNormalizer()
		>>> normalizer.normalized_word('می‌رم')
		['می‌روم', 'می‌رم']
		>>> normalizer = InformalNormalizer(seperation_flag=True)
		>>> normalizer.normalized_word('صداوسیماجمهوری')
		['صداوسیما جمهوری', 'صداوسیماجمهوری']
		"""

		options = []
		if word in self.lemmatizer.words or word in self.lemmatizer.verbs:
			pass

		elif word in self.iverb_map:
			options.append(self.iverb_map[word])

		elif word in self.iword_map:
			options.append(self.iword_map[word])

		elif word[:-2] in self.ilemmatizer.verbs and word.endswith('ین'):
			options.append(word[:-1] + 'د')

		elif word.endswith("ن") and word[:-1] in self.ilemmatizer.verbs:
			options.append(word + 'د')

		elif word[:-1] in self.ilemmatizer.verbs and word.endswith('ه') and word[:-1] not in self.lemmatizer.words:
			options.append(self.iword_map.get(word[:-1], word[:-1]) + 'د')

		elif word not in self.ilemmatizer.verbs and word.endswith('ه') and word[:-1] in self.ilemmatizer.words:
			options.append(self.iword_map.get(word[:-1], word[:-1]) + ' است')

		elif word not in self.ilemmatizer.verbs and word.endswith('ون') and self.lemmatizer.lemmatize(word[:-2] + 'ان') in self.ilemmatizer.words:
			options.append(word[:-2] + 'ان')

		elif self.seperation_flag:
			options.append(self.split_token_words(word))

		options.append(word)
		return options

	def normalize(self, text):

		sent_tokenizer = SentenceTokenizer()
		word_tokenizer = WordTokenizer()
		text = super(InformalNormalizer, self).normalize(text)
		sents = [word_tokenizer.tokenize(sentence) for sentence in sent_tokenizer.tokenize(text)]

		return [[self.normalized_word(word) for word in sent] for sent in sents]

	def informal_conjugations(self, verb):
		ends = ['م', 'ی', '', 'یم', 'ین', 'ن']
		present_simples = [verb + end for end in ends]
		if verb.endswith('ا'):
			present_simples[2] = verb + 'د'
		else:
			present_simples[2] = verb + 'ه'
		present_not_simples = ['ن' + item for item in present_simples]
		present_imperfects = ['می‌' + item for item in present_simples]
		present_not_imperfects = ['ن' + item for item in present_imperfects]
		present_subjunctives = [
			item if item.startswith('ب') else 'ب' + item for item in present_simples]
		present_not_subjunctives = ['ن' + item for item in present_simples]
		return present_simples + present_not_simples + \
			present_imperfects + present_not_imperfects + \
			present_subjunctives + present_not_subjunctives


class InformalLemmatizer(Lemmatizer):

	def __init__(self, **kargs):
		super(InformalLemmatizer, self).__init__(**kargs)

		temp = []
		self.words = set(self.words.keys())
		for word in self.words:
			if word.endswith("ً"):
				temp.append(word[:-1])

		self.words.update(temp)

		temp = {}
		for verb in self.verbs:
			if verb.endswith("د"):
				temp[verb[:-1] + 'ن'] = self.verbs[verb]

		self.verbs.update(temp)

		with codecs.open(informal_verbs, encoding='utf8') as vf:
			for f, i, flag in map(lambda x: x.strip().split(' ', 2), vf):
				self.verbs.update(dict(
					map(lambda x: (x, f), self.iconjugations(i))
				))

		with codecs.open(informal_words, encoding='utf8') as wf:
			self.words.update(
				map(lambda x: x.strip().split(' ', 1)[0], wf)
			)

	def iconjugations(self, verb):
		ends = ['م', 'ی', '', 'یم', 'ین', 'ن']
		present_simples = [verb + end for end in ends]
		if verb.endswith('ا'):
			present_simples[2] = verb + 'د'
		else:
			present_simples[2] = verb + 'ه'
		present_not_simples = ['ن' + item for item in present_simples]
		present_imperfects = ['می‌' + item for item in present_simples]
		present_not_imperfects = ['ن' + item for item in present_imperfects]
		present_subjunctives = [
			item if item.startswith('ب') else 'ب' + item for item in present_simples]
		present_not_subjunctives = ['ن' + item for item in present_simples]
		return present_simples + present_not_simples + \
			present_imperfects + present_not_imperfects + \
			present_subjunctives + present_not_subjunctives



# coding: utf-8

from __future__ import unicode_literals
from .utils import default_words, default_verbs
from .Stemmer import Stemmer
from .WordTokenizer import WordTokenizer


class Lemmatizer(object):
	"""
	>>> lemmatizer = Lemmatizer()
	>>> lemmatizer.lemmatize('کتاب‌ها')
	'کتاب'
	>>> lemmatizer.lemmatize('آتشفشان')
	'آتشفشان'
	>>> lemmatizer.lemmatize('می‌روم')
	'رفت#رو'
	>>> lemmatizer.lemmatize('گفته_شده_است')
	'گفت#گو'
	>>> lemmatizer.lemmatize('نچشیده_است')
	'چشید#چش'
	>>> lemmatizer.lemmatize('مردم', pos='N')
	'مردم'
	>>> lemmatizer.lemmatize('اجتماعی', pos='AJ')
	'اجتماعی'
	"""

	def __init__(self, words_file=default_words, verbs_file=default_verbs, joined_verb_parts=True):
		self.verbs = {}
		self.stemmer = Stemmer()

		tokenizer = WordTokenizer(words_file=default_words, verbs_file=verbs_file)
		self.words = tokenizer.words

		if verbs_file:
			self.verbs['است'] = '#است'
			for verb in tokenizer.verbs:
				for tense in self.conjugations(verb):
					self.verbs[tense] = verb
			if joined_verb_parts:
				for verb in tokenizer.verbs:
					bon = verb.split('#')[0]
					for after_verb in tokenizer.after_verbs:
						self.verbs[bon + 'ه_' + after_verb] = verb
						self.verbs['ن' + bon + 'ه_' + after_verb] = verb
					for before_verb in tokenizer.before_verbs:
						self.verbs[before_verb + '_' + bon] = verb

	def lemmatize(self, word, pos=''):
		if not pos and word in self.words:
			return word

		if (not pos or pos == 'V') and word in self.verbs:
			return self.verbs[word]

		if pos.startswith('AJ') and word[-1] == 'ی':
			return word

		if pos == 'PRO':
			return word

		if word in self.words:
			return word

		stem = self.stemmer.stem(word)
		if stem and stem in self.words:
			return stem

		return word

	def conjugations(self, verb):
		"""
		>>> lemmatizer = Lemmatizer()
		>>> lemmatizer.conjugations('خورد#خور')
		['خوردم', 'خوردی', 'خورد', 'خوردیم', 'خوردید', 'خوردند', 'نخوردم', 'نخوردی', 'نخورد', 'نخوردیم', 'نخوردید', 'نخوردند', 'خورم', 'خوری', 'خورد', 'خوریم', 'خورید', 'خورند', 'نخورم', 'نخوری', 'نخورد', 'نخوریم', 'نخورید', 'نخورند', 'می‌خوردم', 'می‌خوردی', 'می‌خورد', 'می‌خوردیم', 'می‌خوردید', 'می‌خوردند', 'نمی‌خوردم', 'نمی‌خوردی', 'نمی‌خورد', 'نمی‌خوردیم', 'نمی‌خوردید', 'نمی‌خوردند', 'خورده‌ام', 'خورده‌ای', 'خورده', 'خورده‌ایم', 'خورده‌اید', 'خورده‌اند', 'نخورده‌ام', 'نخورده‌ای', 'نخورده', 'نخورده‌ایم', 'نخورده‌اید', 'نخورده‌اند', 'خورم', 'خوری', 'خورد', 'خوریم', 'خورید', 'خورند', 'نخورم', 'نخوری', 'نخورد', 'نخوریم', 'نخورید', 'نخورند', 'می‌خورم', 'می‌خوری', 'می‌خورد', 'می‌خوریم', 'می‌خورید', 'می‌خورند', 'نمی‌خورم', 'نمی‌خوری', 'نمی‌خورد', 'نمی‌خوریم', 'نمی‌خورید', 'نمی‌خورند', 'بخورم', 'بخوری', 'بخورد', 'بخوریم', 'بخورید', 'بخورند', 'نخورم', 'نخوری', 'نخورد', 'نخوریم', 'نخورید', 'نخورند', 'بخور', 'نخور']
		>>> lemmatizer.conjugations('آورد#آور')
		['آوردم', 'آوردی', 'آورد', 'آوردیم', 'آوردید', 'آوردند', 'نیاوردم', 'نیاوردی', 'نیاورد', 'نیاوردیم', 'نیاوردید', 'نیاوردند', 'آورم', 'آوری', 'آورد', 'آوریم', 'آورید', 'آورند', 'نیاورم', 'نیاوری', 'نیاورد', 'نیاوریم', 'نیاورید', 'نیاورند', 'می‌آوردم', 'می‌آوردی', 'می‌آورد', 'می‌آوردیم', 'می‌آوردید', 'می‌آوردند', 'نمی‌آوردم', 'نمی‌آوردی', 'نمی‌آورد', 'نمی‌آوردیم', 'نمی‌آوردید', 'نمی‌آوردند', 'آورده‌ام', 'آورده‌ای', 'آورده', 'آورده‌ایم', 'آورده‌اید', 'آورده‌اند', 'نیاورده‌ام', 'نیاورده‌ای', 'نیاورده', 'نیاورده‌ایم', 'نیاورده‌اید', 'نیاورده‌اند', 'آورم', 'آوری', 'آورد', 'آوریم', 'آورید', 'آورند', 'نیاورم', 'نیاوری', 'نیاورد', 'نیاوریم', 'نیاورید', 'نیاورند', 'می‌آورم', 'می‌آوری', 'می‌آورد', 'می‌آوریم', 'می‌آورید', 'می‌آورند', 'نمی‌آورم', 'نمی‌آوری', 'نمی‌آورد', 'نمی‌آوریم', 'نمی‌آورید', 'نمی‌آورند', 'بیاورم', 'بیاوری', 'بیاورد', 'بیاوریم', 'بیاورید', 'بیاورند', 'نیاورم', 'نیاوری', 'نیاورد', 'نیاوریم', 'نیاورید', 'نیاورند', 'بیاور', 'نیاور']
		"""

		past, present = verb.split('#')
		ends = ['م', 'ی', '', 'یم', 'ید', 'ند']

		if verb == '#هست':
			return ['هست' + end for end in ends] + ['نیست' + end for end in ends]

		past_simples = [past + end for end in ends]
		past_imperfects = ['می‌' + item for item in past_simples]
		ends = ['ه‌ام', 'ه‌ای', 'ه', 'ه‌ایم', 'ه‌اید', 'ه‌اند']
		past_narratives = [past + end for end in ends]

		imperatives = ['ب' + present, 'ن' + present]

		if present.endswith('ا') or present in ('آ', 'گو'):
			present = present + 'ی'

		ends = ['م', 'ی', 'د', 'یم', 'ید', 'ند']
		present_simples = [present + end for end in ends]
		present_imperfects = ['می‌' + item for item in present_simples]
		present_subjunctives = [item if item.startswith('ب') else 'ب' + item for item in present_simples]
		present_not_subjunctives = ['ن' + item for item in present_simples]

		with_nots = lambda items: items + list(map(lambda item: 'ن' + item, items))
		aa_refinement = lambda items: list(map(lambda item: item.replace('بآ', 'بیا').replace('نآ', 'نیا'), items)) if items[0].startswith('آ') else items
		return aa_refinement(with_nots(past_simples) + with_nots(present_simples) + with_nots(past_imperfects) + with_nots(past_narratives) + with_nots(present_simples) + with_nots(present_imperfects) + present_subjunctives + present_not_subjunctives + imperatives)



# coding: utf-8

from __future__ import unicode_literals
from .Lemmatizer import Lemmatizer


class TokenSplitter():
	def __init__(self):
		self.lemmatizer = Lemmatizer()
		self.lemmatize = self.lemmatizer.lemmatize
		self.words = self.lemmatizer.words

	def split_token_words(self, token):
		"""
		>>> splitter = TokenSplitter()
		>>> splitter.split_token_words('صداوسیماجمهوری')
		[('صداوسیما', 'جمهوری')]
		>>> splitter.split_token_words('صداو')
		[('صد', 'او'), ('صدا', 'و')]
		>>> splitter.split_token_words('شهرموشها')
		[('شهر', 'موشها')]
		>>> splitter.split_token_words('داستان‌سرا')
		[('داستان', 'سرا'), ('داستان‌سرا',)]
		>>> splitter.split_token_words('دستان‌سرا')
		[('دستان', 'سرا')]
		"""

		candidates = []
		if '‌' in token:
			candidates.append(tuple(token.split('‌')))

		splits = [(token[:s], token[s:]) for s in range(1, len(token)) if token[s-1] != '‌' and token[s] != '‌'] + [(token, )]
		candidates.extend(list(filter(lambda tokens: set(map(self.lemmatize, tokens)).issubset(self.words), splits)))

		return candidates



# coding: utf-8

from __future__ import unicode_literals
import re
from .Lemmatizer import Lemmatizer
from .WordTokenizer import WordTokenizer
from .utils import maketrans

compile_patterns = lambda patterns: [(re.compile(pattern), repl) for pattern, repl in patterns]


class Normalizer(object):
	def __init__(self, remove_extra_spaces=True, persian_style=True, persian_numbers=True, remove_diacritics=True, affix_spacing=True, token_based=False, punctuation_spacing=True):
		self._punctuation_spacing = punctuation_spacing
		self._affix_spacing = affix_spacing
		self._token_based = token_based

		translation_src, translation_dst = ' كي“”', ' کی""'
		if persian_numbers:
			translation_src += '0123456789%'
			translation_dst += '۰۱۲۳۴۵۶۷۸۹٪'
		self.translations = maketrans(translation_src, translation_dst)

		if self._token_based:
			lemmatizer = Lemmatizer()
			self.words = lemmatizer.words
			self.verbs = lemmatizer.verbs
			self.tokenizer = WordTokenizer(join_verb_parts=False)
			self.suffixes = {'ی', 'ای', 'ها', 'های', 'تر', 'تری', 'ترین', 'گر', 'گری', 'ام', 'ات', 'اش'}

		self.character_refinement_patterns = []

		if remove_extra_spaces:
			self.character_refinement_patterns.extend([
				(r' +', ' '),  # remove extra spaces
				(r'\n\n+', '\n\n'),  # remove extra newlines
				(r'[ـ\r]', ''),  # remove keshide, carriage returns
			])

		if persian_style:
			self.character_refinement_patterns.extend([
				('"([^\n"]+)"', r'«\1»'),  # replace quotation with gyoome
				('([\d+])\.([\d+])', r'\1٫\2'),  # replace dot with momayez
				(r' ?\.\.\.', ' …'),  # replace 3 dots
			])

		if remove_diacritics:
			self.character_refinement_patterns.append(
				('[\u064B\u064C\u064D\u064E\u064F\u0650\u0651\u0652]', ''),  # remove FATHATAN, DAMMATAN, KASRATAN, FATHA, DAMMA, KASRA, SHADDA, SUKUN
			)

		self.character_refinement_patterns = compile_patterns(self.character_refinement_patterns)

		punc_after, punc_before = r'\.:!،؛؟»\]\)\}', r'«\[\(\{'
		if punctuation_spacing:
			self.punctuation_spacing_patterns = compile_patterns([
				('" ([^\n"]+) "', r'"\1"'),  # remove space before and after quotation
				(' (['+ punc_after +'])', r'\1'),  # remove space before
				('(['+ punc_before +']) ', r'\1'),  # remove space after
				('(['+ punc_after[:3] +'])([^ \d'+ punc_after +'])', r'\1 \2'),  # put space after . and :
				('(['+ punc_after[3:] +'])([^ '+ punc_after +'])', r'\1 \2'),  # put space after
				('([^ '+ punc_before +'])(['+ punc_before +'])', r'\1 \2'),  # put space before
			])

		if affix_spacing:
			self.affix_spacing_patterns = compile_patterns([
				(r'([^ ]ه) ی ', r'\1‌ی '),  # fix ی space
				(r'(^| )(ن?می) ', r'\1\2‌'),  # put zwnj after می, نمی
				(r'(?<=[^\n\d '+ punc_after + punc_before +']{2}) (تر(ین?)?|گری?|های?)(?=[ \n'+ punc_after + punc_before +']|$)', r'‌\1'),  # put zwnj before تر, تری, ترین, گر, گری, ها, های
				(r'([^ ]ه) (ا(م|یم|ش|ند|ی|ید|ت))(?=[ \n'+ punc_after +']|$)', r'\1‌\2'),  # join ام, ایم, اش, اند, ای, اید, ات
			])

	def normalize(self, text):
		text = self.character_refinement(text)
		if self._affix_spacing:
			text = self.affix_spacing(text)

		if self._token_based:
			tokens = self.tokenizer.tokenize(text.translate(self.translations))
			text = ' '.join(self.token_spacing(tokens))

		if self._punctuation_spacing:
			text = self.punctuation_spacing(text)

		return text

	def character_refinement(self, text):
		"""
		>>> normalizer = Normalizer()
		>>> normalizer.character_refinement('اصلاح كاف و ياي عربي')
		'اصلاح کاف و یای عربی'

		>>> normalizer.character_refinement('عراق سال 2012 قراردادی به ارزش "4.2 میلیارد دلار" برای خرید تجهیزات نظامی با روسیه امضا  کرد.')
		'عراق سال ۲۰۱۲ قراردادی به ارزش «۴٫۲ میلیارد دلار» برای خرید تجهیزات نظامی با روسیه امضا کرد.'

		>>> normalizer.character_refinement('رمــــان')
		'رمان'

		>>> normalizer.character_refinement('بُشقابِ مَن را بِگیر')
		'بشقاب من را بگیر'
		"""

		text = text.translate(self.translations)
		for pattern, repl in self.character_refinement_patterns:
			text = pattern.sub(repl, text)
		return text

	def punctuation_spacing(self, text):
		"""
		>>> normalizer = Normalizer()
		>>> normalizer.punctuation_spacing('اصلاح ( پرانتزها ) در متن .')
		'اصلاح (پرانتزها) در متن.'

		>>> normalizer.punctuation_spacing('نسخه 0.5 در ساعت 22:00 تهران،1396')
		'نسخه 0.5 در ساعت 22:00 تهران، 1396'
		"""

		for pattern, repl in self.punctuation_spacing_patterns:
			text = pattern.sub(repl, text)
		return text

	def affix_spacing(self, text):
		"""
		>>> normalizer = Normalizer()
		>>> normalizer.affix_spacing('خانه ی پدری')
		'خانه‌ی پدری'

		>>> normalizer.affix_spacing('فاصله میان پیشوند ها و پسوند ها را اصلاح می کند.')
		'فاصله میان پیشوند‌ها و پسوند‌ها را اصلاح می‌کند.'

		>>> normalizer.affix_spacing('می روم')
		'می‌روم'

		>>> normalizer.affix_spacing('حرفه ای')
		'حرفه‌ای'

		>>> normalizer.affix_spacing('محبوب ترین ها')
		'محبوب‌ترین‌ها'
		"""

		for pattern, repl in self.affix_spacing_patterns:
			text = pattern.sub(repl, text)
		return text

	def token_spacing(self, tokens):
		"""
		>>> normalizer = Normalizer(token_based=True)
		>>> normalizer.token_spacing(['کتاب', 'ها'])
		['کتاب‌ها']

		>>> normalizer.token_spacing(['او', 'می', 'رود'])
		['او', 'می‌رود']

		>>> normalizer.token_spacing(['ماه', 'می', 'سال', 'جدید'])
		['ماه', 'می', 'سال', 'جدید']

		>>> normalizer.token_spacing(['اخلال', 'گر'])
		['اخلال‌گر']

		>>> normalizer.token_spacing(['پرداخت', 'شده', 'است'])
		['پرداخت', 'شده', 'است']

		>>> normalizer.token_spacing(['زمین', 'لرزه', 'ای'])
		['زمین‌لرزه‌ای']
		"""

		result = []
		for t, token in enumerate(tokens):
			joined = False

			if result:
				token_pair = result[-1]+'‌'+token
				if token_pair in self.verbs or token_pair in self.words and self.words[token_pair][0] > 0:
					joined = True

					if t < len(tokens)-1 and token+'_'+tokens[t+1] in self.verbs:
						joined = False

				elif token in self.suffixes and result[-1] in self.words:
					joined = True

			if joined:
				result.pop()
				result.append(token_pair)
			else:
				result.append(token)

		return result



# coding: utf-8

from __future__ import print_function
import os, sys, re
from xml.dom import minidom


class TNewsReader():
	"""
	interfaces [TNews Corpus](http://datasets.tnews.ir/downloads/) that you must download and extract.

	>>> tnews = TNewsReader(root='corpora/tnews')
	>>> next(tnews.docs())['id']
	'14092303482300013653'
	"""

	def __init__(self, root):
		self._root = root
		self.cleaner = re.compile(r'<[^<>]+>')

	def docs(self):
		def get_text(element):
			raw_html = element.childNodes[0].data if element.childNodes else ''
			cleaned_text = re.sub(self.cleaner, '', raw_html)
			return cleaned_text

		for root, dirs, files in os.walk(self._root):
			for name in sorted(files):

				try:
					content = open(os.path.join(root, name)).read()

					# fix xml formating issue
					content = re.sub(r'[]', '', content).replace('</TNews>', '') + '</TNews>'

					elements = minidom.parseString(content)
					for element in elements.getElementsByTagName('NEWS'):
						doc = {}
						doc['id'] = get_text(element.getElementsByTagName('NEWSID')[0])
						doc['url'] = get_text(element.getElementsByTagName('URL')[0])
						doc['datetime'] = get_text(element.getElementsByTagName('UTCDATE')[0])
						doc['category'] = get_text(element.getElementsByTagName('CATEGORY')[0])
						doc['pre-title'] = get_text(element.getElementsByTagName('PRETITLE')[0])
						doc['title'] = get_text(element.getElementsByTagName('TITLE')[0])
						doc['post-title'] = get_text(element.getElementsByTagName('POSTTITLE')[0])
						doc['brief'] = get_text(element.getElementsByTagName('BRIEF')[0])
						doc['text'] = get_text(element.getElementsByTagName('DESCRIPTION')[0])
						yield doc

				except Exception as e:
					print('error in reading', name, e, file=sys.stderr)

	def texts(self):
		for doc in self.docs():
			yield doc['text']



# coding: utf-8

import sys, codecs
from os import path

PY2 = sys.version_info[0] == 2

data_path = path.join(path.dirname(__file__), 'data')
default_words = path.join(data_path, 'words.dat')
default_stopwords = path.join(data_path, 'stopwords.dat')
default_verbs = path.join(data_path, 'verbs.dat')
informal_words = path.join(data_path, 'iwords.dat')
informal_verbs = path.join(data_path, 'iverbs.dat')

NUMBERS = '۰۱۲۳۴۵۶۷۸۹'

maketrans = lambda A, B: dict((ord(a), b) for a, b in zip(A, B))


def words_list(words_file=default_words):
	with codecs.open(words_file, encoding='utf-8') as words_file:
		items = [line.strip().split('\t') for line in words_file]
		return [(item[0], int(item[1]), tuple(item[2].split(','))) for item in items if len(item) == 3]


def stopwords_list(stopwords_file=default_stopwords):
	with codecs.open(stopwords_file, encoding='utf8') as stopwords_file:
		return list(map(lambda w: w.strip(), stopwords_file))



# coding: utf-8

from __future__ import unicode_literals
from nltk.stem.api import StemmerI


class Stemmer(StemmerI):
	"""
	>>> stemmer = Stemmer()
	>>> stemmer.stem('کتابی')
	'کتاب'
	>>> stemmer.stem('کتاب‌ها')
	'کتاب'
	>>> stemmer.stem('کتاب‌هایی')
	'کتاب'
	>>> stemmer.stem('کتابهایشان')
	'کتاب'
	>>> stemmer.stem('اندیشه‌اش')
	'اندیشه'
	>>> stemmer.stem('خانۀ')
	'خانه'
	"""

	def __init__(self):
		self.ends = ['ات', 'ان', 'ترین', 'تر', 'م', 'ت', 'ش', 'یی', 'ی', 'ها', 'ٔ', '‌ا', '‌']

	def stem(self, word):
		for end in self.ends:
			if word.endswith(end):
				word = word[:-len(end)]

		if word.endswith('ۀ'):
			word = word[:-1] + 'ه'

		return word



# coding: utf-8

from __future__ import unicode_literals, print_function
import os, re, subprocess


class WikipediaReader():
	"""
	interfaces [Persian Wikipedia dump](http://download.wikimedia.org/fawiki/latest/fawiki-latest-pages-articles.xml.bz2)
	"""

	def __init__(self, fawiki_dump, n_jobs=2):
		self.fawiki_dump = fawiki_dump
		self.wiki_extractor = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'WikiExtractor.py')
		self.n_jobs = n_jobs

	def docs(self):
		proc = subprocess.Popen(['python', self.wiki_extractor, '--no-templates', '--processes', str(self.n_jobs), '--output', '-', self.fawiki_dump], stdout=subprocess.PIPE)
		doc_pattern = re.compile(r'<doc id="(\d+)" url="([^\"]+)" title="([^\"]+)">')

		doc = []
		for line in iter(proc.stdout.readline, ''):
			line = line.strip().decode('utf8')
			if line:
				doc.append(line)

			if line == '</doc>':
				del doc[1]
				id, url, title = doc_pattern.match(doc[0]).groups()
				html = '\n'.join(doc[1:-1])

				yield {'id': id, 'url': url, 'title': title, 'html': html, 'text': html}
				doc = []

	def texts(self):
		for doc in self.docs():
			yield doc['text']



# coding: utf-8

from __future__ import unicode_literals
from nltk.tag.api import TaggerI
from nltk.metrics import accuracy


class SequenceTagger(TaggerI):
	""" wrapper for [Wapiti](http://wapiti.limsi.fr) sequence tagger

	>>> tagger = SequenceTagger(patterns=['*', 'u:word-%x[0,0]'])
	>>> tagger.train([[('من', 'PRO'), ('به', 'P'), ('مدرسه', 'N'), ('رفته_بودم', 'V'), ('.', 'PUNC')]])
	>>> tagger.tag_sents([['من', 'به', 'مدرسه', 'رفته_بودم', '.']])
	[[('من', 'PRO'), ('به', 'P'), ('مدرسه', 'N'), ('رفته_بودم', 'V'), ('.', 'PUNC')]]

	>>> tagger.save_model('resources/test.model')
	>>> SequenceTagger(model='resources/test.model').tag_sents([['من', 'به', 'مدرسه', 'رفته_بودم', '.']])
	[[('من', 'PRO'), ('به', 'P'), ('مدرسه', 'N'), ('رفته_بودم', 'V'), ('.', 'PUNC')]]
	"""

	def __init__(self, patterns=[], **options):
		from wapiti import Model
		self.model = Model(patterns='\n'.join(patterns), **options)

	def train(self, sentences):
		self.model.train(['\n'.join([' '.join(word) for word in sentence]) for sentence in sentences])

	def save_model(self, filename):
		self.model.save(filename)

	def tag(self, tokens):
		return self.tag_sents([tokens])[0]

	def tag_sents(self, sentences):
		sentences = list(sentences)
		lines = '\n\n'.join(['\n'.join(sentence) for sentence in sentences]).replace(' ', '_')
		results = self.model.label_sequence(lines).decode('utf8')
		tags = iter(results.strip().split('\n'))
		return [[(word, next(tags)) for word in sentence] for sentence in sentences]


class IOBTagger(SequenceTagger):
	""" wrapper for [Wapiti](http://wapiti.limsi.fr) sequence tagger

	>>> tagger = IOBTagger(patterns=['*', 'U:word-%x[0,0]', 'U:word-%x[0,1]'])
	>>> tagger.train([[('من', 'PRO', 'B-NP'), ('به', 'P', 'B-PP'), ('مدرسه', 'N', 'B-NP'), ('رفته_بودم', 'V', 'B-VP'), ('.', 'PUNC', 'O')]])
	>>> tagger.tag_sents([[('من', 'PRO'), ('به', 'P'), ('مدرسه', 'N'), ('رفته_بودم', 'V'), ('.', 'PUNC')]])
	[[('من', 'PRO', 'B-NP'), ('به', 'P', 'B-PP'), ('مدرسه', 'N', 'B-NP'), ('رفته_بودم', 'V', 'B-VP'), ('.', 'PUNC', 'O')]]
	"""

	def tag_sents(self, sentences):
		sentences = list(sentences)
		lines = '\n\n'.join(['\n'.join(['\t'.join(word) for word in sentence]) for sentence in sentences]).replace(' ', '_')
		results = self.model.label_sequence(lines).decode('utf8')
		tags = iter(results.strip().split('\n'))
		return [[word + (next(tags),) for word in sentence] for sentence in sentences]

	def evaluate(self, gold):
		tagged_sents = self.tag_sents(([word[:-1] for word in sentence] for sentence in gold))
		return accuracy(sum(gold, []), sum(tagged_sents, []))



# coding: utf-8

from __future__ import unicode_literals
import codecs
from nltk.parse import DependencyGraph
from nltk.tree import Tree


def coarse_pos_e(tags):
	"""
	Coarse POS tags of Dadegan corpus:
		N: Noun, V: Verb, ADJ: Adjective, ADV: Adverb, PR: Pronoun, PREP: Preposition, POSTP: Postposition, CONJ: Conjunction, PUNC: Punctuation, ADR: Address Term, IDEN: Title, PART: Particle, POSNUM: Post-noun Modifier, PREM: Pre-modifier, PRENUM: Pre-noun Numeral, PSUS: Pseudo-sentence, SUBR: Subordinating Clause

	>>> coarse_pos_e(['N', 'IANM'])
	'N'
	"""

	map = {'N': 'N', 'V': 'V', 'ADJ': 'AJ', 'ADV': 'ADV', 'PR': 'PRO', 'PREM': 'DET', 'PREP': 'P', 'POSTP': 'POSTP', 'PRENUM': 'NUM', 'CONJ': 'CONJ', 'PUNC': 'PUNC', 'SUBR': 'CONJ'}
	return map.get(tags[0], 'X') + ('e' if 'EZ' in tags else '')


word_nodes = lambda tree: sorted(tree.nodes.values(), key=lambda node: node['address'])[1:]
node_deps = lambda node: sum(node['deps'].values(), [])


class DadeganReader():
	"""
	interfaces [Persian Dependency Treebank](http://dadegan.ir/perdt/download)

	>>> dadegan = DadeganReader(conll_file='corpora/dadegan.conll')
	>>> next(dadegan.sents())
	[('این', 'DET'), ('میهمانی', 'N'), ('به', 'P'), ('منظور', 'Ne'), ('آشنایی', 'Ne'), ('هم‌تیمی‌های', 'Ne'), ('او', 'PRO'), ('با', 'P'), ('غذاهای', 'Ne'), ('ایرانی', 'AJ'), ('ترتیب', 'N'), ('داده_شد', 'V'), ('.', 'PUNC')]

	>>> from hazm.Chunker import tree2brackets
	>>> tree2brackets(next(dadegan.chunked_trees()))
	'[این میهمانی NP] [به PP] [منظور آشنایی هم‌تیمی‌های او NP] [با PP] [غذاهای ایرانی NP] [ترتیب داده_شد VP] .'
	"""

	def __init__(self, conll_file, pos_map=coarse_pos_e):
		self._conll_file = conll_file
		self._pos_map = pos_map if pos_map else lambda tags: ','.join(tags)

	def _sentences(self):
		with codecs.open(self._conll_file, encoding='utf8') as conll_file:
			text = conll_file.read()

			# refine text
			text = text.replace('‌‌', '‌').replace('\t‌', '\t').replace('‌\t', '\t').replace('\t ', '\t').replace(' \t', '\t').replace(
				'\r', '').replace('\u2029', '‌')

			for item in text.replace(' ', '_').split('\n\n'):
				if item.strip():
					yield item

	def trees(self):
		for sentence in self._sentences():
			tree = DependencyGraph(sentence)

			for node in word_nodes(tree):
				node['mtag'] = [node['ctag'], node['tag']]

				if 'ezafe' in node['feats']:
					node['mtag'].append('EZ')

				node['mtag'] = self._pos_map(node['mtag'])

			yield tree

	def sents(self):
		for tree in self.trees():
			yield [(node['word'], node['mtag']) for node in word_nodes(tree)]

	def chunked_trees(self):
		for tree in self.trees():
			chunks = []
			for node in word_nodes(tree):
				n = node['address']
				item = (node['word'], node['mtag'])
				appended = False
				if node['ctag'] in {'PREP', 'POSTP'}:
					for d in node_deps(node):
						label = 'PP'
						if node['ctag'] == 'POSTP':
							label = 'POSTP'
						if d == n - 1 and type(chunks[-1]) == Tree and chunks[-1].label() == label:
							chunks[-1].append(item)
							appended = True
					if node['head'] == n - 1 and len(chunks) > 0 and type(chunks[-1]) == Tree and chunks[
						-1].label() == label:
						chunks[-1].append(item)
						appended = True
					if not appended:
						chunks.append(Tree(label, [item]))
				elif node['ctag'] in {'PUNC', 'CONJ', 'SUBR', 'PART'}:
					if item[0] in {"'", '"', '(', ')', '{', '}', '[', ']', '-', '#', '«', '»'} and len(chunks) > 0 and type(chunks[-1]) == Tree:
						for l in chunks[-1].leaves():
							if l[1] == item[1]:
								chunks[-1].append(item)
								appended = True
								break
					if appended is not True:
						chunks.append(item)
				elif node['ctag'] in {'N', 'PREM', 'ADJ', 'PR', 'ADR', 'PRENUM', 'IDEN', 'POSNUM', 'SADV'}:
					if node['rel'] in {'MOZ', 'NPOSTMOD'}:
						if len(chunks) > 0:
							if type(chunks[-1]) == Tree:
								j = n - len(chunks[-1].leaves())
								chunks[-1].append(item)
							else:
								j = n - 1
								treeNode = Tree('NP', [chunks.pop(), item])
								chunks.append(treeNode)
							while j > node['head']:
								leaves = chunks.pop().leaves()
								if len(chunks) < 1:
									chunks.append(Tree('NP', leaves))
									j -= 1
								elif type(chunks[-1]) == Tree:
									j -= len(chunks[-1])
									for l in leaves:
										chunks[-1].append(l)
								else:
									leaves.insert(0, chunks.pop())
									chunks.append(Tree('NP', leaves))
									j -= 1
							continue
					elif node['rel'] == 'POSDEP' and tree.nodes[node['head']]['rel'] in {'NCONJ', 'AJCONJ'}:
						conj = tree.nodes[node['head']]
						if tree.nodes[conj['head']]['rel'] in {'MOZ', 'NPOSTMOD', 'AJCONJ', 'POSDEP'}:
							label = 'NP'
							leaves = [item]
							j = n - 1
							while j >= conj['head']:
								if type(chunks[-1]) is Tree:
									j -= len(chunks[-1].leaves())
									label = chunks[-1].label()
									leaves = chunks.pop().leaves() + leaves
								else:
									leaves.insert(0, chunks.pop())
									j -= 1
							chunks.append(Tree(label, leaves))
							appended = True
					elif node['head'] == n - 1 and len(chunks) > 0 and type(chunks[-1]) == Tree and not chunks[
						-1].label() == 'PP':
						chunks[-1].append(item)
						appended = True
					elif node['rel'] == 'AJCONJ' and tree.nodes[node['head']]['rel'] in {'NPOSTMOD', 'AJCONJ'}:
						np_nodes = [item]
						label = 'ADJP'
						i = n - node['head']
						while i > 0:
							if type(chunks[-1]) == Tree:
								label = chunks[-1].label()
								leaves = chunks.pop().leaves()
								i -= len(leaves)
								np_nodes = leaves + np_nodes
							else:
								i -= 1
								np_nodes.insert(0, chunks.pop())
						chunks.append(Tree(label, np_nodes))
						appended = True
					elif node['ctag'] == 'ADJ' and node['rel'] == 'POSDEP' and tree.nodes[node['head']]['ctag'] != 'CONJ':
						np_nodes = [item]
						i = n - node['head']
						while i > 0:
							label = 'ADJP'
							if type(chunks[-1]) == Tree:
								label = chunks[-1].label()
								leaves = chunks.pop().leaves()
								i -= len(leaves)
								np_nodes = leaves + np_nodes
							else:
								i -= 1
								np_nodes.insert(0, chunks.pop())
						chunks.append(Tree(label, np_nodes))
						appended = True
					for d in node_deps(node):
						if d == n - 1 and type(chunks[-1]) == Tree and chunks[
							-1].label() != 'PP' and appended is not True:
							label = chunks[-1].label()
							if node['rel'] == 'ADV':
								label = 'ADVP'
							elif label in {'ADJP', 'ADVP'}:
								if node['ctag'] == 'N':
									label = 'NP'
								elif node['ctag'] == 'ADJ':
									label = 'ADJP'
							leaves = chunks.pop().leaves()
							leaves.append(item)
							chunks.append(Tree(label, leaves))
							appended = True
						elif tree.nodes[d]['rel'] == 'NPREMOD' and appended is not True:
							np_nodes = [item]
							i = n - d
							while i > 0:
								if type(chunks[-1]) == Tree:
									leaves = chunks.pop().leaves()
									i -= len(leaves)
									np_nodes = leaves + np_nodes
								else:
									i -= 1
									np_nodes.insert(0, chunks.pop())
							chunks.append(Tree('NP', np_nodes))
							appended = True
					if not appended:
						label = 'NP'
						if node['ctag'] == 'ADJ':
							label = 'ADJP'
						elif node['rel'] == 'ADV':
							label = 'ADVP'
						chunks.append(Tree(label, [item]))
				elif node['ctag'] in {'V'}:
					appended = False
					for d in node_deps(node):
						if d == n - 1 and type(chunks[-1]) == Tree and tree.nodes[d]['rel'] in {'NVE', 'ENC'} and appended is not True:
							leaves = chunks.pop().leaves()
							leaves.append(item)
							chunks.append(Tree('VP', leaves))
							appended = True
						elif tree.nodes[d]['rel'] in {'VPRT', 'NVE'}:
							vp_nodes = [item]
							i = n - d
							while i > 0:
								if type(chunks[-1]) == Tree:
									leaves = chunks.pop().leaves()
									i -= len(leaves)
									vp_nodes = leaves + vp_nodes
								else:
									i -= 1
									vp_nodes.insert(0, chunks.pop())
							chunks.append(Tree('VP', vp_nodes))
							appended = True
							break
					if not appended:
						chunks.append(Tree('VP', [item]))
				elif node['ctag'] in {'PSUS'}:
					if node['rel'] == 'ADV':
						chunks.append(Tree('ADVP', [item]))
					else:
						chunks.append(Tree('VP', [item]))
				elif node['ctag'] in {'ADV', 'SADV'}:
					appended = False
					for d in node_deps(node):
						if d == n - 1 and type(chunks[-1]) == Tree:
							leaves = chunks.pop().leaves()
							leaves.append(item)
							chunks.append(Tree('ADVP', leaves))
							appended = True
					if not appended:
						chunks.append(Tree('ADVP', [item]))

			yield Tree('S', chunks)



# coding: utf-8

from __future__ import unicode_literals
import os, codecs
from .Normalizer import Normalizer
from .WordTokenizer import WordTokenizer


def coarse_pos_e(tags):
	"""
	Coarse POS tags of Peykare corpus:
		N: Noun, V: Verb, AJ: Adjective, ADV: Adverb, PRO: Pronoun, DET: Determiner, P: Preposition, POSTP: Postposition, NUM: Number, CONJ: Conjunction, PUNC: Punctuation, RES: Residual, CL: Classifier, INT: Interjection

	>>> coarse_pos_e(['N','COM','SING'])
	'N'
	"""

	try:
		return list(set(tags) & {'N', 'V', 'AJ', 'ADV', 'PRO', 'DET', 'P', 'POSTP', 'NUM', 'CONJ', 'PUNC', 'CL', 'INT', 'RES'})[0] + ('e' if 'EZ' in tags else '')
	except:
		return 'N'


def join_verb_parts(sentence):
	"""
	Join verb parts like Dadedgan corpus.

	>>> join_verb_parts([('اولین', 'AJ'), ('سیاره', 'Ne'), ('خارج', 'AJ'), ('از', 'P'), ('منظومه', 'Ne'), ('شمسی', 'AJ'), ('دیده', 'AJ'), ('شد', 'V'), ('.', 'PUNC')])
	[('اولین', 'AJ'), ('سیاره', 'Ne'), ('خارج', 'AJ'), ('از', 'P'), ('منظومه', 'Ne'), ('شمسی', 'AJ'), ('دیده_شد', 'V'), ('.', 'PUNC')]
	"""

	if not hasattr(join_verb_parts, 'tokenizer'):
		join_verb_parts.tokenizer = WordTokenizer()
	before_verbs, after_verbs, verbe = join_verb_parts.tokenizer.before_verbs, join_verb_parts.tokenizer.after_verbs, join_verb_parts.tokenizer.verbe

	result = [('', '')]
	for word in reversed(sentence):
		if word[0] in before_verbs or (result[-1][0] in after_verbs and word[0] in verbe):
			result[-1] = (word[0] + '_' + result[-1][0], result[-1][1])
		else:
			result.append(word)
	return list(reversed(result[1:]))


class PeykareReader():
	"""
	Interfaces [Peykare Corpus](http://www.rcisp.com/?q=%D9%BE%DB%8C%DA%A9%D8%B1%D9%87%20%D9%85%D8%AA%D9%86%DB%8C%20%D8%B2%D8%A8%D8%A7%D9%86%20%D9%81%D8%A7%D8%B1%D8%B3%DB%8C)
	Bijankhan, M., Sheykhzadegan, J., Bahrani, M., & Ghayoomi, M. (2011). Lessons from building a Persian written corpus: Peykare. Language Resources and Evaluation, 45, 143–164.

	>>> peykare = PeykareReader(root='corpora/peykare')
	>>> next(peykare.sents())
	[('دیرزمانی', 'N'), ('از', 'P'), ('راه‌اندازی', 'Ne'), ('شبکه‌ی', 'Ne'), ('خبر', 'Ne'), ('الجزیره', 'N'), ('نمی‌گذرد', 'V'), ('،', 'PUNC'), ('اما', 'CONJ'), ('این', 'DET'), ('شبکه‌ی', 'Ne'), ('خبری', 'AJe'), ('عربی', 'N'), ('بسیار', 'ADV'), ('سریع', 'ADV'), ('توانسته', 'V'), ('در', 'P'), ('میان', 'Ne'), ('شبکه‌های', 'Ne'), ('عظیم', 'AJe'), ('خبری', 'AJ'), ('و', 'CONJ'), ('بنگاه‌های', 'Ne'), ('چندرسانه‌ای', 'AJe'), ('دنیا', 'N'), ('خودی', 'N'), ('نشان', 'N'), ('دهد', 'V'), ('.', 'PUNC')]

	Reading Peykare sentences without mapping pos tags:

	>>> peykare = PeykareReader(root='corpora/peykare', joined_verb_parts=False, pos_map=None)
	>>> next(peykare.sents())
	[('دیرزمانی', 'N,COM,SING,TIME,YA'), ('از', 'P'), ('راه‌اندازی', 'N,COM,SING,EZ'), ('شبکه‌ی', 'N,COM,SING,EZ'), ('خبر', 'N,COM,SING,EZ'), ('الجزیره', 'N,PR,SING'), ('نمی‌گذرد', 'V,PRES,NEG,3'), ('،', 'PUNC'), ('اما', 'CONJ'), ('این', 'DET,DEMO'), ('شبکه‌ی', 'N,COM,SING,EZ'), ('خبری', 'AJ,SIM,EZ'), ('عربی', 'N,PR,SING'), ('بسیار', 'ADV,INTSF,SIM'), ('سریع', 'ADV,GENR,SIM'), ('توانسته', 'V,PASTP'), ('در', 'P'), ('میان', 'N,COM,SING,EZ'), ('شبکه‌های', 'N,COM,PL,EZ'), ('عظیم', 'AJ,SIM,EZ'), ('خبری', 'AJ,SIM'), ('و', 'CONJ'), ('بنگاه‌های', 'N,COM,PL,EZ'), ('چندرسانه‌ای', 'AJ,SIM,EZ'), ('دنیا', 'N,COM,SING'), ('خودی', 'N,COM,SING,YA'), ('نشان', 'N,COM,SING'), ('دهد', 'V,SUB,POS,3'), ('.', 'PUNC')]
	"""

	def __init__(self, root, joined_verb_parts=True, pos_map=coarse_pos_e):
		self._root = root
		self._pos_map = pos_map if pos_map else lambda tags: ','.join(tags)
		self._joined_verb_parts = joined_verb_parts
		self._normalizer = Normalizer(punctuation_spacing=False, affix_spacing=False)

	def docs(self):
		""" extracts raw text of peykare document """

		for root, dirs, files in os.walk(self._root):
			for name in sorted(files):
				with codecs.open(os.path.join(root, name), encoding='windows-1256') as peykare_file:
					text = peykare_file.read()
					if text:
						yield text

	def doc_to_sents(self, document):
		""" converts extracted document text to a list of (word, tag) pairs """

		sentence = []
		for line in document.split('\r\n'):
			if not line:
				continue

			parts = line.split(' ')
			tags, word = parts[3], self._normalizer.normalize('‌'.join(parts[4:]))

			if word and word != '#':
				sentence.append((word, tags))

			if parts[2] == 'PUNC' and word in {'#', '.', '؟', '!'}:
				if len(sentence) > 1:
					yield sentence
				sentence = []

	def sents(self):
		map_pos = lambda item: (item[0], self._pos_map(item[1].split(',')))

		for document in self.docs():
			for sentence in self.doc_to_sents(document):
				if self._joined_verb_parts:
					sentence = join_verb_parts(sentence)

				yield list(map(map_pos, sentence))



# coding: utf-8

from __future__ import unicode_literals
from nltk.tag import stanford
from .SequenceTagger import SequenceTagger


class POSTagger(SequenceTagger):
	"""
	>>> tagger = POSTagger(model='resources/postagger.model')
	>>> tagger.tag(['من', 'به', 'مدرسه', 'رفته_بودم', '.'])
	[('من', 'PRO'), ('به', 'P'), ('مدرسه', 'N'), ('رفته_بودم', 'V'), ('.', 'PUNC')]
	"""


class StanfordPOSTagger(stanford.StanfordPOSTagger):
	"""
	>>> tagger = StanfordPOSTagger(model_filename='resources/persian.tagger', path_to_jar='resources/stanford-postagger.jar')
	>>> tagger.tag(['من', 'به', 'مدرسه', 'رفته_بودم', '.'])
	[('من', 'PRO'), ('به', 'P'), ('مدرسه', 'N'), ('رفته_بودم', 'V'), ('.', 'PUNC')]
	"""

	def __init__(self, model_filename, path_to_jar, *args, **kwargs):
		self._SEPARATOR = '/'
		super(stanford.StanfordPOSTagger, self).__init__(model_filename=model_filename, path_to_jar=path_to_jar, *args, **kwargs)

	def tag(self, tokens):
		return self.tag_sents([tokens])[0]

	def tag_sents(self, sentences):
		refined = map(lambda s: [w.replace(' ', '_') for w in s], sentences)
		return super(stanford.StanfordPOSTagger, self).tag_sents(refined)



# coding: utf-8

from __future__ import print_function
import os, sys, re
from xml.dom import minidom


class HamshahriReader():
	"""
	interfaces [Hamshahri Corpus](http://dbrg.ut.ac.ir/Hamshahri/download.html#version2) that you must download and extract it.

	>>> hamshahri = HamshahriReader(root='corpora/hamshahri')
	>>> next(hamshahri.docs())['id']
	'HAM2-750403-001'
	"""

	def __init__(self, root):
		self._root = root
		self._invalids = set(['hamshahri.dtd', 'HAM2-960622.xml', 'HAM2-960630.xml', 'HAM2-960701.xml', 'HAM2-960709.xml', 'HAM2-960710.xml', 'HAM2-960711.xml', 'HAM2-960817.xml', 'HAM2-960818.xml', 'HAM2-960819.xml', 'HAM2-960820.xml', 'HAM2-961019.xml', 'HAM2-961112.xml', 'HAM2-961113.xml', 'HAM2-961114.xml', 'HAM2-970414.xml', 'HAM2-970415.xml', 'HAM2-970612.xml', 'HAM2-970614.xml', 'HAM2-970710.xml', 'HAM2-970712.xml', 'HAM2-970713.xml', 'HAM2-970717.xml', 'HAM2-970719.xml', 'HAM2-980317.xml', 'HAM2-040820.xml', 'HAM2-040824.xml', 'HAM2-040825.xml', 'HAM2-040901.xml', 'HAM2-040917.xml', 'HAM2-040918.xml', 'HAM2-040920.xml', 'HAM2-041025.xml', 'HAM2-041026.xml', 'HAM2-041027.xml', 'HAM2-041230.xml', 'HAM2-041231.xml', 'HAM2-050101.xml', 'HAM2-050102.xml', 'HAM2-050223.xml', 'HAM2-050224.xml', 'HAM2-050406.xml', 'HAM2-050407.xml', 'HAM2-050416.xml'])
		self._paragraph_pattern = re.compile(r'(\n.{0,50})(?=\n)')

	def docs(self):
		for root, dirs, files in os.walk(self._root):
			for name in sorted(files):
				if name in self._invalids:
					continue

				try:
					elements = minidom.parse(os.path.join(root, name))
					for element in elements.getElementsByTagName('DOC'):
						doc = {}
						doc['id'] = element.getElementsByTagName('DOCID')[0].childNodes[0].data
						doc['issue'] = element.getElementsByTagName('ISSUE')[0].childNodes[0].data

						for cat in element.getElementsByTagName('CAT'):
							doc['categories_'+ cat.attributes['xml:lang'].value] = cat.childNodes[0].data.split('.')

						for date in element.getElementsByTagName('DATE'):
							if date.attributes['calender'].value == 'Persian':
								doc['date'] = date.childNodes[0].data

						elm = element.getElementsByTagName('TITLE')[0]
						doc['title'] = elm.childNodes[1].data if len(elm.childNodes) > 1 else ''

						doc['text'] = ''
						for item in element.getElementsByTagName('TEXT')[0].childNodes:
							if item.nodeType == 4:  # CDATA
								doc['text'] += item.data

						# refine text
						doc['text'] = self._paragraph_pattern.sub(r'\1\n', doc['text']).replace('\no ', '\n')

						yield doc

				except Exception as e:
					print('error in reading', name, e, file=sys.stderr)

	def texts(self):
		for doc in self.docs():
			yield doc['text']



# coding: utf-8

from __future__ import unicode_literals, print_function
import os, sys, re, codecs
from xml.dom import minidom
from nltk.tree import Tree
from .WordTokenizer import WordTokenizer


def coarse_pos_e(tags):
	"""
	Coarse POS tags of Treebank corpus:
		N: Noun, V: Verb, A: Adjective, D: Adverb, Z: Pronoun, T: Determiner, E: Preposition, P: Postposition, U: Number, J: Conjunction, O: Punctuation, R: Residual, L: Classifier, I: Interjection

	>>> coarse_pos_e(['Nasp---', 'pers', 'prop'])
	'N'
	"""

	map = {'N': 'N', 'V': 'V', 'A': 'AJ', 'D': 'ADV', 'Z': 'PRO', 'T': 'DET', 'E': 'P', 'P': 'POSTP', 'U': 'NUM', 'J': 'CONJ', 'O': 'PUNC', 'R': 'RES', 'L': 'CL', 'I': 'INT'}
	try:
		if tags[0][0] == 'C':
			if 'pronominal' in tags:
				tags[0] = 'Z'
			elif 'verb' in tags:
				tags[0] = 'V'
			elif 'prep' in tags:
				tags[0] = 'E'
			elif 'adv' in tags:
				tags[0] = 'D'
			elif 'det' in tags:
				tags[0] = 'T'
		return map[tags[0][0]] + ('e' if 'ezafe' in tags else '')
	except Exception:
		return ''


class TreebankReader():
	"""
	interfaces [Per­si­an Tree­bank](http://hpsg.fu-berlin.de/~ghayoomi/PTB.html)

	>>> treebank = TreebankReader(root='corpora/treebank')
	>>> print(next(treebank.trees()))
	(S
	  (VPS
	    (NPC (N دنیای/Ne) (MN (N آدولف/N) (N بورن/N)))
	    (VPC
	      (NPC (N دنیای/Ne) (NPA (N اتفاقات/Ne) (ADJ رویایی/AJ)))
	      (V است/V)))
	  (PUNC ./PUNC))

	>>> next(treebank.sents())
	[('دنیای', 'Ne'), ('آدولف', 'N'), ('بورن', 'N'), ('دنیای', 'Ne'), ('اتفاقات', 'Ne'), ('رویایی', 'AJ'), ('است', 'V'), ('.', 'PUNC')]

	>>> from .Chunker import tree2brackets
	>>> tree2brackets(next(treebank.chunked_trees()))
	'[دنیای آدولف بورن NP] [دنیای اتفاقات رویایی NP] [است VP] .'
	"""

	def __init__(self, root, pos_map=coarse_pos_e, join_clitics=False, join_verb_parts=False):
		self._root = root
		self._pos_map = pos_map if pos_map else lambda tags: ','.join(tags)
		self._join_clitics = join_clitics
		self._join_verb_parts = join_verb_parts
		self._tokenizer = WordTokenizer()

	def docs(self):
		for root, dirs, files in os.walk(self._root):
			for name in sorted(files):
				try:
					with codecs.open(os.path.join(root, name), encoding='utf8') as treebank_file:
						raw = re.sub(r'\n *', '', treebank_file.read())
						yield minidom.parseString(raw.encode('utf8'))
				except Exception as e:
					print('error in reading', name, e, file=sys.stderr)

	def trees(self):

		def traverse(node):
			def extract_tags(W):
				pos = [W.getAttribute('lc') if W.getAttribute('lc') else None]
				if W.getAttribute('clitic') in {'ezafe', 'pronominal', 'verb', 'prep', 'adv', 'det'}:
					pos.append(W.getAttribute('clitic'))
				if W.getAttribute('ne_sort'):
					pos.append(W.getAttribute('ne_sort'))
				if W.getAttribute('n_type'):
					pos.append(W.getAttribute('n_type'))
				if W.getAttribute('ya_type'):
					pos.append(W.getAttribute('ya_type'))
				if W.getAttribute('ke_type'):
					pos.append(W.getAttribute('ke_type'))
				if W.getAttribute('type'):
					pos.append(W.getAttribute('type'))
				if W.getAttribute('kind'):
					pos.append(W.getAttribute('kind'))
				return pos

			def clitic_join(tree, clitic):
				if type(tree[-1]) == Tree:
					return clitic_join(tree[-1], clitic)
				else:
					if(clitic[0][0][0] == 'ا'):
						clitic[0] = ('‌' + clitic[0][0], clitic[0][1])
					tree[-1]=(tree[-1][0] + clitic[0][0], clitic[0][1])
					tree.set_label('CLITICS')
					return

			if not len(node.childNodes):
				return
			first = node.childNodes[0]
			if first.tagName == 'w':
				pos=extract_tags(first)
				return Tree(node.tagName, [(first.childNodes[0].data.replace('می ', 'می‌'), self._pos_map(pos))])
			childs = node.childNodes[2:] if node.tagName == 'S' else node.childNodes
			for child in childs:
				if not len(child.childNodes):
					childs.remove(child)
			tree = Tree(node.tagName, map(traverse, childs))
			if self._join_clitics and len(tree) > 1 and type(tree[1]) == Tree and tree[1].label() == 'CLITIC' and tree[1][0][1] not in {'P', 'V'}:
				clitic=tree[-1]
				tree = Tree(tree.label(), [subtree for subtree in tree[0]])
				clitic_join(tree, clitic)
			if self._join_verb_parts and len(tree) > 1 and type(tree[1]) == Tree and type(tree[0]) == Tree and tree[0].label() == 'AUX' and tree[0][0][0] in self._tokenizer.before_verbs:
				tree[1][0] = (tree[0][0][0] + ' ' + tree[1][0][0], tree[1][0][1])
				tree.remove(tree[0])
			if self._join_verb_parts and len(tree.leaves()) > 1 and tree.leaves()[-1][0] in self._tokenizer.after_verbs and tree.leaves()[-2][0] in self._tokenizer.verbe :
				tree[1][0] = (tree[0].leaves()[-1][0] + ' ' + tree[1][0][0], tree[1][0][1])
				path = tree.leaf_treeposition(len(tree.leaves())-2)
				removingtree = tree
				while len(path) > 2 :
					removingtree = removingtree[path[0]]
					path = path[1:]
				removingtree.remove(Tree(tree.pos()[-2][1],[tree.pos()[-2][0]]))
			if self._join_verb_parts and len(tree.leaves()) > 1 and tree.leaves()[-1][0] in self._tokenizer.after_verbs and tree.leaves()[-2][0] in self._tokenizer.verbe :
				tree[1][0] = (tree[0].leaves()[-1][0] + ' ' + tree[1][0][0], tree[1][0][1])
				path = tree.leaf_treeposition(len(tree.leaves())-2)
				removingtree = tree
				while len(path) > 2 :
					removingtree = removingtree[path[0]]
					path = path[1:]
				removingtree.remove(Tree(tree.pos()[-2][1],[tree.pos()[-2][0]]))
			return tree

		for doc in self.docs():
			for S in doc.getElementsByTagName('S'):
				yield traverse(S)

	def sents(self):
		for tree in self.trees():
			yield tree.leaves()


	def chunked_trees(self):
		collapse = lambda node, label: Tree(label, [Tree(pos[1], [pos[0]]) for pos in node.pos()])

		def traverse(node, parent, chunks):
			label = node.label()

			if label.count('-nid') > 0:
				label = label.replace('-nid', '')
			if label.count('-nid') > 0:
				label = label.replace('-nid', '')
			if label.count('-DiscA') > 0:
				label = label.replace('-DiscA', '')

			if label in {'CLITIC', 'CLITICS'}:
				if node[0][1] == 'V':
					label = 'V'
				elif node[0][1] == 'P':
					label = 'PREP'
				elif node[0][1] == 'DET':
					label = 'DET'
				elif node[0][1] == 'ADV':
					label = 'ADV'
				elif node[0][1] == 'PRO':
					label = 'PRON'

			if label in {'CONJ', 'PUNC'} and len(node) == 1:
				chunks.append(node)
				return

			if label == 'PPC' and len(node) == 1:
				chunks.append(Tree('PP', [node[0]]))
				return

			if label == 'PREP':
				chunks.append(Tree('PP', [node]))
				return

			if label == 'PostP':
				chunks.append(Tree('POSTP', [node]))
				return

			for leaf in node.pos():
				if leaf[1] in {'PUNC', 'CONJ', 'PREP', 'PostP'}:
					for i in range(len(node)):
						traverse(node[i], node, chunks)
					return

			if label == 'NPA' and parent.label() in {'CPC', 'PPC'}:
				chunks.append(collapse(node, 'NP'))
				return

			if label == 'NPA' and len(node)>=1:
				if node[0].label() == 'ADV':
					chunks.append(collapse(node, 'NP'))
					return

			if label in {'NPC', 'N', 'INFV', 'DPA', 'CLASS', 'DPC', 'DEM', 'INTJ', 'MN', 'PRON', 'DET', 'NUM', 'RES'}:
				chunks.append(collapse(node, 'NP'))
				return

			if label == 'NPA' and len(node) >= 2:
				if node[0].label() == 'ADJ' and node[1].label() == 'NPC' or node[0].label() in {'N', 'PRON'} and node[1].label() in {'ADJ', 'ADJPA', 'N'} or node[0].label() == 'NUM' and node[1].label() in {'N', 'NPC', 'MN', 'NUM'} or node[0].label() in {'N', 'NPC', 'MN'} and node[1].label() == 'NUM' or node[0].label() == 'NPC' and node[1].label() == 'ADJ' or node[0].label() == 'NPA' and node[1].label() != 'NPC' or node[1].label() == 'NPA' and node[0].label() != 'NPC':
					chunks.append(collapse(node, 'NP'))
					return

			if label == 'DPC' and len(node) >= 2:
				chunkable = True
				for leaf in node[1].pos():
					if leaf[1] in {'PUNC', 'CONJ', 'PREP', 'PostP'}:
						chunkable = False
				if node[1].label() in {'N', 'NPA', 'NPC'} and chunkable:
					chunks.append(collapse(node, 'NP'))
					return

			if label == 'DPA' and len(node)>=2:
				if node[1].label() == 'ADV':
					chunks.append(collapse(node, 'ADVP'))
					return

			if label in {'MV', 'V', 'AUX', 'PPARV'}:
				chunks.append(Tree('VP', [node]))
				return

			if label in {'ADJ', 'ADJPC', 'MADJ', 'ADVPA'}:
				chunks.append(Tree('ADJP', [node]))
				return

			if label in {'ADV', 'MADV', 'ADVPC'}:
				chunks.append(Tree('ADVP', [node]))
				return

			if type(node[0]) != Tree:
				chunks.append(node)
				return

			for i in range(len(node)):
				traverse(node[i], node, chunks)

		for tree in self.trees():
			chunks = []
			traverse(tree, None, chunks)
			for i in range(len(chunks)):
				if chunks[i].label() in {'PUNC', 'CONJ'}:
					chunks[i] = chunks[i][0]
				else:
					chunks[i] = Tree(chunks[i].label(), chunks[i].leaves())
			yield Tree('S', chunks)



# coding: utf-8

from __future__ import unicode_literals
import re, codecs
from .utils import words_list, default_words, default_verbs
from nltk.tokenize.api import TokenizerI


class WordTokenizer(TokenizerI):
	"""
	>>> tokenizer = WordTokenizer()
	>>> tokenizer.tokenize('این جمله (خیلی) پیچیده نیست!!!')
	['این', 'جمله', '(', 'خیلی', ')', 'پیچیده', 'نیست', '!!!']

	>>> tokenizer.tokenize('نسخه 0.5 در ساعت 22:00 تهران،1396')
	['نسخه', '0.5', 'در', 'ساعت', '22:00', 'تهران', '،', '1396']

	>>> tokenizer = WordTokenizer(join_verb_parts=False)
	>>> print(' '.join(tokenizer.tokenize('سلام.')))
	سلام .

	>>> tokenizer = WordTokenizer(join_verb_parts=False, replace_links=True)
	>>> print(' '.join(tokenizer.tokenize('در قطر هک شد https://t.co/tZOurPSXzi https://t.co/vtJtwsRebP')))
	در قطر هک شد LINK LINK

	>>> tokenizer = WordTokenizer(join_verb_parts=False, replace_IDs=True, replace_numbers=True)
	>>> print(' '.join(tokenizer.tokenize('زلزله ۴.۸ ریشتری در هجدک کرمان @bourse24ir')))
	زلزله NUMF ریشتری در هجدک کرمان ID

	>>> tokenizer = WordTokenizer(join_verb_parts=False, replace_hashtags=True, replace_numbers=True, separate_emoji=True)
	>>> print(' '.join(tokenizer.tokenize('📍عرضه بلوک 17 درصدی #های_وب به قیمت')))
	📍 عرضه بلوک NUM2 درصدی TAG های وب به قیمت

	>>> tokenizer = WordTokenizer(join_verb_parts=False, separate_emoji=True)
	>>> print(' '.join(tokenizer.tokenize('دیگه میخوام ترک تحصیل کنم 😂😂😂')))
	دیگه میخوام ترک تحصیل کنم 😂 😂 😂
	"""

	def __init__(self, words_file=default_words, verbs_file=default_verbs, join_verb_parts=True, separate_emoji=False, replace_links=False, replace_IDs=False, replace_emails=False, replace_numbers=False, replace_hashtags=False):
		self._join_verb_parts = join_verb_parts
		self.separate_emoji = separate_emoji
		self.replace_links = replace_links
		self.replace_IDs = replace_IDs
		self.replace_emails = replace_emails
		self.replace_numbers = replace_numbers
		self.replace_hashtags = replace_hashtags

		self.pattern = re.compile(r'([؟!\?]+|\d[\d\.:/\\]+|[:\.،؛»\]\)\}"«\[\(\{])')  # TODO \d
		self.emoji_pattern = re.compile(u"["
										u"\U0001F600-\U0001F64F"  # emoticons
										u"\U0001F300-\U0001F5FF"  # symbols & pictographs
										u"\U0001F4CC\U0001F4CD"  # pushpin & round pushpin
										"]", flags=re.UNICODE)
		self.emoji_repl = r'\g<0> '
		self.id_pattern = re.compile(r'([^\w\._]+)(@[\w_]+)')
		self.id_repl = r'\1ID'
		self.link_pattern = re.compile(r'((https?|ftp):\/\/)?(?<!@)([wW]{3}\.)?(([\w-]+)(\.(\w){2,})+([-\w@:%_\+\/~#?&]+)?)')
		self.link_repl = r'LINK'
		self.email_pattern = re.compile(r'[a-zA-Z0-9\._\+-]+@([a-zA-Z0-9-]+\.)+[A-Za-z]{2,}')
		self.email_repl = r'EMAIL'
		self.number_int_pattern = re.compile(r'([^\.,\w]+)([\d۰-۹]+)([^\.,\w]+)')
		self.number_int_repl = lambda m: m.group(1) + 'NUM'+ str(len(m.group(2))) + m.group(3)
		self.number_float_pattern = re.compile(r'([^,\w]+)([\d۰-۹,]+[\.٫]{1}[\d۰-۹]+)([^,\w]+)')
		self.number_float_repl = r'\1NUMF\3'
		self.hashtag_pattern = re.compile(r'\#([\S]+)')
		# NOTE: python2.7 does not support unicodes with \w  Example: r'\#([\w\_]+)'

		self.hashtag_repl = lambda m: 'TAG ' + m.group(1).replace('_', ' ')

		self.words = {item[0]: (item[1], item[2]) for item in words_list(default_words)}

		if join_verb_parts:
			self.after_verbs = set([
				'ام', 'ای', 'است', 'ایم', 'اید', 'اند', 'بودم', 'بودی', 'بود', 'بودیم', 'بودید', 'بودند', 'باشم', 'باشی', 'باشد', 'باشیم', 'باشید', 'باشند',
				'شده_ام', 'شده_ای', 'شده_است', 'شده_ایم', 'شده_اید', 'شده_اند', 'شده_بودم', 'شده_بودی', 'شده_بود', 'شده_بودیم', 'شده_بودید', 'شده_بودند', 'شده_باشم', 'شده_باشی', 'شده_باشد', 'شده_باشیم', 'شده_باشید', 'شده_باشند',
				'نشده_ام', 'نشده_ای', 'نشده_است', 'نشده_ایم', 'نشده_اید', 'نشده_اند', 'نشده_بودم', 'نشده_بودی', 'نشده_بود', 'نشده_بودیم', 'نشده_بودید', 'نشده_بودند', 'نشده_باشم', 'نشده_باشی', 'نشده_باشد', 'نشده_باشیم', 'نشده_باشید', 'نشده_باشند',
				'شوم', 'شوی', 'شود', 'شویم', 'شوید', 'شوند', 'شدم', 'شدی', 'شد', 'شدیم', 'شدید', 'شدند',
				'نشوم', 'نشوی', 'نشود', 'نشویم', 'نشوید', 'نشوند', 'نشدم', 'نشدی', 'نشد', 'نشدیم', 'نشدید', 'نشدند',
				'می‌شوم', 'می‌شوی', 'می‌شود', 'می‌شویم', 'می‌شوید', 'می‌شوند', 'می‌شدم', 'می‌شدی', 'می‌شد', 'می‌شدیم', 'می‌شدید', 'می‌شدند',
				'نمی‌شوم', 'نمی‌شوی', 'نمی‌شود', 'نمی‌شویم', 'نمی‌شوید', 'نمی‌شوند', 'نمی‌شدم', 'نمی‌شدی', 'نمی‌شد', 'نمی‌شدیم', 'نمی‌شدید', 'نمی‌شدند',
				'خواهم_شد', 'خواهی_شد', 'خواهد_شد', 'خواهیم_شد', 'خواهید_شد', 'خواهند_شد',
				'نخواهم_شد', 'نخواهی_شد', 'نخواهد_شد', 'نخواهیم_شد', 'نخواهید_شد', 'نخواهند_شد',
			])

			self.before_verbs = set([
				'خواهم', 'خواهی', 'خواهد', 'خواهیم', 'خواهید', 'خواهند',
				'نخواهم', 'نخواهی', 'نخواهد', 'نخواهیم', 'نخواهید', 'نخواهند'
			])

			with codecs.open(verbs_file, encoding='utf8') as verbs_file:
				self.verbs = list(reversed([verb.strip() for verb in verbs_file if verb]))
				self.bons = set([verb.split('#')[0] for verb in self.verbs])
				self.verbe = set([bon +'ه' for bon in self.bons] + ['ن'+ bon +'ه' for bon in self.bons])

	def tokenize(self, text):

		if self.separate_emoji:
			text = self.emoji_pattern.sub(self.emoji_repl, text)
		if self.replace_links:
			text = self.link_pattern.sub(self.link_repl, text)
		if self.replace_IDs:
			text = self.id_pattern.sub(self.id_repl, text)
		if self.replace_emails:
			text = self.email_pattern.sub(self.email_repl, text)
		if self.replace_hashtags:
			text = self.hashtag_pattern.sub(self.hashtag_repl, text)
		if self.replace_numbers:
			text = self.number_int_pattern.sub(self.number_int_repl, text)
			text = self.number_float_pattern.sub(self.number_float_repl, text)

		text = self.pattern.sub(r' \1 ', text.replace('\n', ' ').replace('\t', ' '))

		tokens = [word for word in text.split(' ') if word]
		if self._join_verb_parts:
			tokens = self.join_verb_parts(tokens)
		return tokens

	def join_verb_parts(self, tokens):
		"""
		>>> tokenizer = WordTokenizer()
		>>> tokenizer.join_verb_parts(['خواهد', 'رفت'])
		['خواهد_رفت']
		>>> tokenizer.join_verb_parts(['رفته', 'است'])
		['رفته_است']
		>>> tokenizer.join_verb_parts(['گفته', 'شده', 'است'])
		['گفته_شده_است']
		>>> tokenizer.join_verb_parts(['گفته', 'خواهد', 'شد'])
		['گفته_خواهد_شد']
		>>> tokenizer.join_verb_parts(['خسته', 'شدید'])
		['خسته', 'شدید']
		"""

		result = ['']
		for token in reversed(tokens):
			if token in self.before_verbs or (result[-1] in self.after_verbs and token in self.verbe):
				result[-1] = token +'_'+ result[-1]
			else:
				result.append(token)
		return list(reversed(result[1:]))



# coding: utf-8

from __future__ import print_function
import codecs


class PersicaReader():
	"""
	interfaces [Persica Corpus](https://sourceforge.net/projects/persica/)

	>>> persica = PersicaReader('corpora/persica.csv')
	>>> next(persica.docs())['id']
	843656
	"""

	def __init__(self, csv_file):
		self._csv_file = csv_file

	def docs(self):
		lines = []
		for line in codecs.open(self._csv_file, encoding='utf-8-sig'):
			line = line.strip()
			if line:
				if line.endswith(','):
					lines.append(line[:-1])
				else:
					lines.append(line)
					yield {
						'id': int(lines[0]),
						'title': lines[1],
						'text': lines[2],
						'date': lines[3],
						'time': lines[4],
						'category': lines[5],
						'category2': lines[6],
					}
					lines = []

	def texts(self):
		for doc in self.docs():
			yield doc['text']




import codecs
from os import path
from setuptools import setup


setup(
	name='hazm',
	version='0.7.0',
	description='Python library for digesting Persian text.',
	author='Alireza Nourian',
	author_email='az.nourian@gmail.com',
	url='http://www.sobhe.ir/hazm/',
	long_description=codecs.open(path.join(path.abspath(path.dirname(__file__)), 'README.md'), encoding='utf-8').read(),
	long_description_content_type='text/markdown',
	packages=['hazm'],
	package_data={'hazm': ['data/*.dat']},
	classifiers=[
		'Topic :: Text Processing',
		'Natural Language :: Persian',
		'Programming Language :: Python :: 2.7',
		'Programming Language :: Python :: 3.4',
		'Programming Language :: Python :: 3.5',
		'Programming Language :: Python :: 3.6',
		'License :: OSI Approved :: MIT License',
	],
	install_requires=['nltk==3.3', 'libwapiti>=0.2.1;platform_system!="Windows"'],
	extras_require={'wapiti': ['libwapiti>=0.2.1']},
)
